import java.awt.*;

public class Camera {
	double w,h;
	Poly2d[] zlist;
	int znum;
	//============================================
	//constructor
	//============================================
	public Camera(int width, int height) {
		this.w = (double) width;
		this.h = (double) height;
		this.znum = 0;
		this.zlist = new Poly2d[1024];
	}
	//============================================
	//scan polygon from a modellist and get zlist of 2d polygons
	//============================================
	public void scan(Zone zone) {
		this.zreset();
		for (int j = 0; j < zone.num; j++) {
			zone.list[j].absolute();
			double tmp = zone.list[j].center.dist();	//backface culling
			for (int i = 0 ; i < zone.list[j].pnum; i++) {
				Poly2d p = new Poly2d( zone.list[j].plist[i] , zone.list[j].points, this.w, this.h);
				if (p.z <= tmp || zone.list[j].plist[j].backface_culling == false) {
					this.zappend(p);
				}
			}
			zone.list[j].relative();
		}
		this.zsort();
	}
	//============================================
	//draw all polygons from a zlist
	//============================================
	public void draw(Graphics g) {
		for (int i = 0; i < this.znum; i++){
			g.setColor(this.zlist[i].fill);	//interior color of the polygon
			g.fillPolygon(this.zlist[i].x, this.zlist[i].y, this.zlist[i].num);
			g.setColor(this.zlist[i].draw);	//border color of the polygon
			g.drawPolygon(this.zlist[i].x, this.zlist[i].y, this.zlist[i].num);
		}
	}
	//================================================================================================================
	//zlist
	//================================================================================================================
	public void zreset() {
		this.znum = 0;
	}

	public void zappend(Poly2d p) {
		//it does not allow avoidable polygon
		//and protect it self from overflow
		if (this.znum < this.zlist.length && p.avoid == false) {
			this.zlist[this.znum] = p;
			this.znum += 1;
		}
	}

	public void zsort() {
		//sort all 2d polygon from the fartest
		//gnome sorting of all polygon from fartest to nearest
		int pivot = 0;
		while (pivot < this.znum) {
			if ( pivot == 0 || this.zlist[pivot-1].z >= this.zlist[pivot].z ) {
				pivot += 1;
			}
			else {
					Poly2d tmp = this.zlist[pivot];
					this.zlist[pivot] = this.zlist[pivot-1];
					this.zlist[pivot-1] = tmp;
					pivot -= 1;
			}
		}
	}
	//================================================================================================================
	//zlist
	//================================================================================================================
	//================================================================================================================
	//2d polygon
	//================================================================================================================
	public class Poly2d {
		int[] x;
		int[] y;
		double z;
		int num;
		Color fill;
		Color draw;
		boolean avoid;		//make sure this flag is false before adding it to draw list

		public Poly2d(Model.Poly p, PointList points, double scale_w, double scale_h) {
			this.avoid = true;		//first consider we work for nothing, so everything can be canceled by a simple return
			this.x = new int[p.num];		//java polygon use this kind of structure
			this.y = new int[p.num];		//int[] x, int[] y, int n;
			Point center = new Point();	//the center is the average of all point in the poly
			int tmpc = 0 ;			//offscreen point counter
			Point tmpP;			//temporary point make code more lisible
			for (this.num = 0; this.num < p.num; this.num++) {
				tmpP = points.list[ p.list[this.num]];
				//avoid polygon behind		(bug fix)
				if (tmpP.z <= 0){
					return;		//if one point of the polygon is behind
							//then the process end here
							//this is the place for a polyclip algo
				}
				//transform 3d to 2d
				this.x[this.num] = (int)((scale_w * tmpP.x / tmpP.z) + scale_w)/2;	//magic perspective calculation x = x/z , y = y/z
				this.y[this.num] = (int)(((scale_w * tmpP.y / tmpP.z) + scale_h)/2);	//also convert from cartesian to computer screen
				//calculate the center of polygon by the way	(zsort)
				center.add(tmpP);	//simply the average of all point
				//avoid polygon completely offscreen  (acceleration)
				if (this.x[this.num] > scale_w || this.x[this.num] < 0 || this.y[this.num] > scale_h || this.y[this.num] < 0) {
					tmpc += 1;	//count every point offscreen
				}
			}
			//avoid polygon completely offscreen
			if (tmpc == this.num) {
				return;		//if all point of the polygon are offscreen , it end here
			}
			//calculate center of polygon
			center.div((double)p.num);	//here we finish the average of all point
			//calculate distance for zsort and backface culling
			this.z = center.dist();		//distance from center of the polygon and 0,0,0(camera)
			//color
			this.fill = p.color;
			this.draw = p.border;
			this.avoid = false;	//now we allow the polygone to be part of the zlist
		}
	}
	//================================================================================================================
	//2d polygon
	//================================================================================================================
	//============================================
	//end
	//============================================
}
